@@ -9,7 +9,18 @@
 
 /* rx_user.c contains routines specific to the user space UNIX implementation of rx */
 
-# include <afs/param.h>
+/* This controls the size of an fd_set; it must be defined early before
+ * the system headers define that type and the macros that operate on it.
+ * Its value should be as large as the maximum file descriptor limit we
+ * are likely to run into on any platform.  Right now, that is 65536
+ * which is the default hard fd limit on Solaris 9 */
+#define FD_SETSIZE 65536
+ 
+#include <afsconfig.h>
+#include <afs/param.h>
+
+RCSID("$Header: /cvs/openafs/src/rx/rx_lwp.c,v 1.10.4.2 2004/10/18 05:52:25 shadow Exp $");
+
 # include <sys/types.h>		/* fd_set on older platforms */
 # include <errno.h>
 # include <signal.h>
@@ -58,7 +69,7 @@ void rxi_Wakeup(void *addr)
 }
 
 PROCESS rx_listenerPid = 0;	/* LWP process id of socket listener process */
-void rx_ListenerProc(void *dummy);
+static void rx_ListenerProc(void *dummy);
 
 /*
  * Delay the current thread the specified number of seconds.
@@ -149,7 +160,10 @@ void rxi_ListenerProc(rfds, tnop, newcallp)
     afs_int32 nextPollTime;		/* time to next poll FD before sleeping */
     int lastPollWorked, doingPoll;	/* true iff last poll was useful */
     struct timeval tv, *tvp;
-    int i, code;
+    int code;
+#ifdef AFS_NT40_ENV
+    int i;
+#endif
     PROCESS pid;
     char name[MAXTHREADNAMELENGTH] = "srv_0";
 
@@ -424,7 +438,9 @@ int rxi_Sendmsg(osi_socket socket, struct msghdr *msg_p, int flags)
 #elif defined(AFS_LINUX22_ENV)
 	  /* linux unfortunately returns ECONNREFUSED if the target port
 	   * is no longer in use */
-	if (errno != EWOULDBLOCK && errno != ENOBUFS && errno != ECONNREFUSED)
+	  /* and EAGAIN if a UDP checksum is incorrect */
+	if (errno != EWOULDBLOCK && errno != ENOBUFS &&
+	    errno != ECONNREFUSED && errno != EAGAIN)
 #else
 	if (errno != EWOULDBLOCK && errno != ENOBUFS)
 #endif
@@ -440,4 +456,5 @@ int rxi_Sendmsg(osi_socket socket, struct msghdr *msg_p, int flags)
     }
     if (sfds)
 	IOMGR_FreeFDSet(sfds);
+    return 0;
 }
